/*
 * AutoDAO - Generic DAO on steroids implementation for Java.
 *
 * Copyright 2008-2010  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.autodao.impl;

import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.sf.autodao.Dao;
import net.sf.autodao.Finder;
import net.sf.autodao.PersistentEntity;
import net.sf.autodao.QueryArgumentTransformer;
import org.aopalliance.aop.Advice;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.aop.framework.ProxyFactoryBean;
import org.springframework.beans.factory.BeanFactory;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Required;
import org.springframework.core.annotation.AnnotationUtils;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionDefinition;
import org.springframework.transaction.interceptor.DefaultTransactionAttribute;
import org.springframework.transaction.interceptor.MatchAlwaysTransactionAttributeSource;
import org.springframework.transaction.interceptor.TransactionInterceptor;
import org.springframework.util.Assert;

/**
 * Base ORM-agnostic DAO factory.
 * <p/>
 * NOT FOR PUBLIC USE.
 */
public abstract class AbstractDaoFactoryBean extends ProxyFactoryBean implements InitializingBean {
  private static final String QUERY_ARGUMENT_TRANSFORMERS_BEAN_NAME = "autodaoQueryArgumentTransformers";
  private static final String LEGACY_QUERY_ARGUMENT_TRANSFORMER_BEAN_NAME = "autodaoQueryArgumentTransformer";
  private static final Logger LOG = LoggerFactory.getLogger(AbstractDaoFactoryBean.class);

  @NotNull
  static final Map<Class<?>, QueryArgumentTransformer> DEFAULT_TRANSFORMERS = defaultTransformers();

  @NotNull
  private static Map<Class<?>, QueryArgumentTransformer> defaultTransformers() {
    final Map<Class<?>, QueryArgumentTransformer> result = new HashMap<Class<?>, QueryArgumentTransformer>();

    ReadableInstantTransformer.register(result);

    return Collections.unmodifiableMap(result);
  }

  @NotNull
  private Map<Class<?>, QueryArgumentTransformer> transformers = DEFAULT_TRANSFORMERS;
  @NotNull
  private Class<?> daoInterface;
  @NotNull
  private PlatformTransactionManager transactionManager;
  @Nullable
  private Class<? extends PersistentEntity<?>> entityType;

  @NotNull
  protected final Map<Class<?>, QueryArgumentTransformer> getTransformers() {
    return transformers;
  }

  @NotNull
  protected abstract FinderExecutor createDao(@NotNull final Class<?> daoInterface,
                                              @Nullable final Class<?> entityType);

  protected abstract void validate(@NotNull final Class<?> daoInterface);

  @Required
  public final void setDaoInterface(@NotNull final Class<?> daoInterface) {
    this.daoInterface = daoInterface;
  }

  @Required
  public final void setTransactionManager(@NotNull final PlatformTransactionManager transactionManager) {
    this.transactionManager = transactionManager;
  }

  public final void setEntityType(@Nullable final Class<? extends PersistentEntity<?>> entityType) {
    this.entityType = entityType;
  }

  @Override
  @SuppressWarnings({"unchecked"})
  public final void setBeanFactory(final BeanFactory beanFactory) {
    transformers = new HashMap<Class<?>, QueryArgumentTransformer>();
    transformers.putAll(DEFAULT_TRANSFORMERS);
    if (beanFactory.containsBean(QUERY_ARGUMENT_TRANSFORMERS_BEAN_NAME)) {
      transformers.putAll((Map<Class<?>, QueryArgumentTransformer>) beanFactory.getBean(QUERY_ARGUMENT_TRANSFORMERS_BEAN_NAME, Map.class));
    }
    if (beanFactory.containsBean(LEGACY_QUERY_ARGUMENT_TRANSFORMER_BEAN_NAME)) {
      LOG.warn(String.format("'%s' bean is no longer supported. Please, consult AutoDAO documentation.", LEGACY_QUERY_ARGUMENT_TRANSFORMER_BEAN_NAME));
    }
    super.setBeanFactory(beanFactory);
  }

  @Override
  public final void afterPropertiesSet() {
    validate(daoInterface);
    final Class<? extends PersistentEntity<?>> realEntityType = getEntityType(daoInterface, entityType);
    addInterface(daoInterface);
    setTarget(createDao(daoInterface, realEntityType));
    addAdvice(createTransactionInterceptor(transactionManager));
    addAdvice(new FinderIntroductionInterceptor(transformers));
  }

  @NotNull
  private static Advice createTransactionInterceptor(@NotNull final PlatformTransactionManager transactionManager) {
    final MatchAlwaysTransactionAttributeSource source = new MatchAlwaysTransactionAttributeSource();
    source.setTransactionAttribute(new DefaultTransactionAttribute(TransactionDefinition.PROPAGATION_MANDATORY));
    return new TransactionInterceptor(transactionManager, source);
  }

  @Nullable
  private static Class<? extends PersistentEntity<?>> getEntityType(@NotNull final Class<?> daoInterface,
                                                                    @Nullable final Class<? extends PersistentEntity<?>> explicit) {
    final Class<? extends PersistentEntity<?>> result;
    if (explicit == null && Dao.class.isAssignableFrom(daoInterface)) {
      result = autoDiscoverEntityType(daoInterface);
      Assert.notNull(result, "No entity type found for " + daoInterface);
    } else {
      result = explicit;
    }
    if (result != null) {
      Assert.isAssignable(PersistentEntity.class, result, "Invalid entity type: ");
    }
    return result;
  }

  @SuppressWarnings({"unchecked"})
  private static Class<? extends PersistentEntity<?>> autoDiscoverEntityType(final Class<?> daoInterface) {
    //Automagically discover entity type from generics.
    for (final Type t : daoInterface.getGenericInterfaces()) {
      if (t instanceof ParameterizedType) {
        final ParameterizedType pt = (ParameterizedType) t;
        if (pt.getRawType() == Dao.class) {
          final Type type = pt.getActualTypeArguments()[0];
          if (type instanceof ParameterizedType) {
            //Special handling for cases like Dao<Entity<Foo>>.
            return (Class<? extends PersistentEntity<?>>) ((ParameterizedType) type).getRawType();
          } else {
            Assert.isInstanceOf(
                Class.class,
                type,
                "Don't know how to handle " + type + " entity type for " + daoInterface.getName() + ".\n"
                    + "Please update to latest AutoDAO release.\n"
                    + "If that doesn't help, please file a bugreport to AutoDAO bugtracker.");
            return (Class<? extends PersistentEntity<?>>) type;
          }
        }
      }
    }
    return null;
  }

  /**
   * @param m method to validate.
   * @return <code>true</code> if method is single-find, <code>false</code> otherwise.
   * @throws SpecificationViolationException
   *
   */
  protected boolean validateReturnType(@NotNull final Method m) {
    final Finder f = AnnotationUtils.findAnnotation(m, Finder.class);
    final boolean singleFind = !Collection.class.isAssignableFrom(m.getReturnType());
    if (!singleFind) {
      final Class<?> returnType = m.getReturnType();
      if (!returnType.isAssignableFrom(f.returnAs())) {
        throw new SpecificationViolationException(
            "@Finder.returnAs value (" + f.returnAs().getName()
                + ") cannot be cast to method return type (" + returnType.getName() + ")", m
        );
      } else if (f.returnAs() != List.class) {
        if (f.returnAs().isInterface()) {
          throw new SpecificationViolationException(
              "@Finder.returnAs value (" + f.returnAs().getName() + ") cannot be an interface", m
          );
        } else {
          try {
            f.returnAs().getConstructor();
          } catch (NoSuchMethodException e) {
            throw new SpecificationViolationException(
                "@Finder.returnAs value (" + f.returnAs().getName() + ") doesn't have no-arg constructor.", m
            );
          }
        }
      }
    }
    return singleFind;
  }
}
